
(*

	PCX (16,256,65536) Loader/Saver unit v1.00 
	   by Ruzaev Sergei ( druidman@chat.ru )

	for Graph unit of TMT Pascal

		PCX ( With LineStop-RLE Code ) :

	16 VGA COLORS load to 256,65536 modes
      256 SVGA COLORS load to 256,65536 modes
    65536 SVGA COLORS load to 65536 mode only

      256/65536 colors modes avail for SavePCX

*)

unit Form_PCX;

interface

const
	(*	Position for Encoding	*)
	center		= 1;
	normal		= 0;

	(*	Errors			*)
	allok		= 0;
	badfile		= 1;
	notsupport	= 2;
	badpage		= 3;

type
	oneRGB=record
	r,g,b			: byte;
	end;

	PCXHeader=record
	manufact		: byte;
	version			: byte;
	encod			: byte;
	bitpix			: byte;
	x0,y0,x1,y1		: word;
	hRes,vRes		: word;
	colorMap		: array[0..47] of byte;
	reserved		: byte;
	nPlanes			: byte;
	bytesperline		: word;
	palInfo			: word;
	filler			: array[0..58] of byte;
	end;

var
PCXError		: byte;
RGBPal			: array[0..767] of byte;	-- real RGB table

function  PCXErrorSTR: string;
procedure LoadPCX(filename: string; position: byte);
procedure SavePCX(filename: string; xw0,yw0,xw1,yw1: word);

implementation

uses Graph;

-- get a color number from a bits
function get16colornum(v,r,g,b: byte): word; assembler;
asm
	XOR	DH,DH
	MOV	DL,V
	SHL	DL,1
	ADD	DL,R
	SHL	DL,1
	ADD	DL,G
	SHL	DL,1
	ADD	DL,B
	MOV	AX,DX
end;

-- get a bit from a byte
function getbit(i,j: byte): word; assembler;
asm
	MOV	DL,I
	MOV	CL,J
	SHR	DL,CL
	SHL	DL,CL
	MOV	CL,7
	SUB	CL,J
	SHL	DL,CL
	SHR	DL,CL
	MOV	CL,J
	SHR	DL,CL
	XOR	DH,DH
	MOV	AX,DX
end;

-- 256 color to 65536 color ungrade
function tabulate(r,g,b: byte): word; assembler;
asm
	MOV	DH,R
	SHR	DH,1
	MOV	DL,G
	SHL	DL,2
	SHL	DX,3
	MOV	CL,B
	SHR	CL,1
	ADD	DL,CL
	MOV	AX,DX
end;

-- 16 bit color to R,G,B incs
procedure GetRGBIncs(cNum: word;var r,g,b: word); assembler;
asm
	mov	dx,cNum				-- r value
	shr	dx,11
	shl	dx,1
	mov	esi,r
	mov	[esi],dx
	mov	dx,cNum				-- g value
	shl	dx,5
	shr	dx,10
	mov	esi,g
	mov	[esi],dx
	mov	dx,cNum				-- b value
	shl	dx,11
	shr	dx,10
	mov	esi,b
	mov	[esi],dx
end;

function PCXErrorStr: string;
begin
case PCXError of
 allok		: result := 'All Ok';
 badfile	: result := 'Bad PCX file';
 notsupport	: result := 'This type of PCX isn''t support';
 badpage	: result := 'Bad Graph settings';
end;
end;

procedure LoadPCX(filename: string; position: byte);
var
f		: file;
thisHead	: PCXHeader;
rgbtab		: array[0..255] of word;	-- tab for 16 mode
maxcurcolor	: longint;			-- max cur color
i,n,z,x,y	: word;
xk0,yk0		: word;
xk1,yk1		: word;

procedure _enc16;			-- PCX 16 color decoder

type
	grad=record
	r,g,b,v	: byte;
	end;

var
cash		: array[0..999] of byte;
wr,ws		: longint;
planes		: byte;
bitplane	: array[0..4095] of grad;
xC,iC		: word;
pr,pg,pb,pv	: byte;				-- pushes (!)

procedure loadCash;
begin
blockread(f,cash,1000,ws); wr:=0;		-- loads up (!)
end;

begin
loadCash;
for y:=yk0 to yk1 do				-- y loop
 begin
 for planes:=0 to 3 do
 begin
 x:=xk0;
 while x<(xk1+1) do				-- x loop
  begin
  i:=cash[wr]; inc(wr);				-- take from cash
  if wr=ws then loadCash;
  if i>192 then
   begin
   n:=i and $3f;
   i:=cash[wr]; inc(wr);
   if wr=ws then loadCash;
   for z:=1 to n do
    begin
    case planes of
    0: bitplane[x].r:=i;
    1: bitplane[x].g:=i;
    2: bitplane[x].b:=i;
    3: bitplane[x].v:=i;
    end;
    inc(x,8);
    end;
   end else
   begin
    case planes of
    0: bitplane[x].r:=i;
    1: bitplane[x].g:=i;
    2: bitplane[x].b:=i;
    3: bitplane[x].v:=i;
    end;
   inc(x,8);
   end;
  end;
 end;
 xC:=xk0;
 while xC<=xk1 do
  begin
  with bitPlane[xC] do
   begin
   pr:=r; pg:=g; pb:=b; pv:=v;
   for iC:=0 to 7 do
    begin
    r:=getbit(pr,iC);
    g:=getbit(pg,iC);
    b:=getbit(pb,iC);
    v:=getbit(pv,iC);
    putpixel(xC+7-iC,y,rgbtab[get16colornum(v,b,g,r)]);
    end;
   end;
  inc(xC,8);
  end;
 end;
end;

procedure _enc256;			-- PCX 256 color decoder
var
cash		: array[0..999] of byte;
wr,ws		: longint;

procedure loadCash;
begin
blockread(f,cash,1000,ws); wr:=0;		-- loads up (!)
end;

begin
loadCash;
for y:=yk0 to yk1 do				-- y loop
 begin
 x:=xk0;
 while x<(xk1+1) do				-- x loop
  begin
  i:=cash[wr]; inc(wr);				-- take from cash
  if wr=ws then loadCash;
  if i>192 then
   begin
   n:=i and $3f;
   i:=cash[wr]; inc(wr);
   if wr=ws then loadCash;
   for z:=1 to n do
    begin
    putpixel(x,y,rgbtab[i]);
    inc(x);
    end;
   end else
   begin
   putpixel(x,y,rgbtab[i]);
   inc(x);
   end;
  end;
 end;
end;

procedure _enc65536;			-- PCX 65536 color decoder
var
cash		: array[0..999] of byte;
wr,ws		: longint;
plane		: byte;
pcolor		: array[0..4095] of oneRGB;
xC		: word;

procedure loadCash;
begin
blockread(f,cash,1000,ws); wr:=0;		-- loads up (!)
end;

begin
loadCash;
for y:=yk0 to yk1 do				-- y loop
 begin
 for plane:=0 to 2 do
 begin
 x:=xk0;
 while x<(xk1+1) do				-- x loop
  begin
  i:=cash[wr]; inc(wr);				-- take from cash
  if wr=ws then loadCash;
  if i>192 then
   begin
   n:=i and $3f;
   i:=cash[wr]; inc(wr);
   if wr=ws then loadCash;
   for z:=1 to n do
    begin
    case plane of
     0: pcolor[x].r:=i shr 2;
     1: pcolor[x].g:=i shr 2;
     2: pcolor[x].b:=i shr 2;
    end;
    inc(x);
    end;
   end else
   begin
   case plane of
    0: pcolor[x].r:=i shr 2;
    1: pcolor[x].g:=i shr 2;
    2: pcolor[x].b:=i shr 2;
   end;
   inc(x);
   end;
  end;
 end;
 for xC:=xk0 to xk1 do
  with pcolor[xC] do
   putpixel(xC,y,tabulate(r,g,b));
 end;
end;

begin

	maxcurcolor:=getmaxcolor;	-- take max color
	if maxcurcolor>65535 then	-- bad page mode (?)
 	 begin
	 PCXError:=badpage;
	 exit;
	 end;


	assign(f,filename);		-- open PCX file
  {$i-} reset(f,1); {$i+}
	if ioresult<>0 then		-- check io
	 begin
	 PCXError:=badfile;
	 exit;
	 end;

	blockread(f,thisHead,128);	-- load PCX header
	
	with thisHead do		-- check support
	begin
	if (bitpix<>1)  and
	   (bitpix<>8)  and
	   (bitpix<>16) then
	 begin
  	 close(f);
	 PCXError:=notsupport;
	 exit;
	 end;
	end;

	if thisHead.bitpix=1 then	-- 16 color pcx
	with thisHead do		-- upgrade to 256 mode
	begin
	for i:=0 to 47 do		-- copy only
	 rgbpal[i]:=colorMap[i];
	end;

	if (thisHead.bitpix=8) and	-- 256 color palette read
	   (thisHead.nPlanes=1) then
	begin
	seek(f,filesize(f)-768);
	blockread(f,rgbpal,768);
	seek(f,128);
	end;

	for i:=0 to 767 do		-- HLS -> RGB
	rgbpal[i]:=rgbpal[i] shr 2;

	for i:=0 to 255 do		-- Fill RGB tab table
	rgbtab[i]:=i;

	if ((thisHead.bitpix=8) and	-- 256 to 65536 mode
	   (thisHead.nplanes<3) and 
	   (maxcurcolor>256)) or
	   ((thisHead.bitpix=1) and
	   (maxcurcolor>256)) then
	begin
	for i:=0 to 255 do		-- real upgrade
	 rgbtab[i]:=tabulate
		(rgbpal[i*3+0],
		 rgbpal[i*3+1],
		 rgbpal[i*3+2]);
	end;

	with thisHead do		-- make coords
	begin
	case position of
	 normal : begin
		  xk0:=x0; yk0:=y0;
		  xk1:=x1; yk1:=y1;
		  end;
	 center : begin
		  if (x1-x0)<getmaxX then
		   begin
		   xk0:=getmaxX div 2-(x1-x0) div 2;
		   xk1:=xk0+(x1-x0);
		   end else
		   begin
                   xk0:=x0;
		   xk1:=x1;
		   end;
		  if (y1-y0)<getmaxY then
		   begin
		   yk0:=getmaxY div 2-(y1-y0) div 2;
		   yk1:=yk0+(y1-y0);
		   end else
		   begin
		   yk0:=y0;
		   yk1:=y1;
		   end;
		  end;
	end;
	end;

	case thisHead.bitpix of		-- start decode
	  1: _enc16;
	  8: case thisHead.nPlanes of
	     1: _enc256;
	     3: _enc65536;
	     end;
	end;


	close(f);			-- all done with ok
	PCXError:=allok;
end;

procedure SavePCX(filename: string; xw0,yw0,xw1,yw1: word);

const

	pcxMagicByte=12;		-- HLS table :))
	j1=$c0;				-- magic bytes (!) 
	j2=$c0 xor $ff;

var
f		: file;
thisHead	: PCXHeader;
maxColor	: longint;
tmpPal		: array[0..767] of byte;
i		: word;

procedure _enc256;			-- pcx 256 encoder
var
cash		: array[0..999] of byte;
wr		: longint;
x,xGo,y		: word;
i1,i2		: byte;

procedure SaveCash;			-- save bf to disk
begin
blockwrite(f,cash,wr); wr:=0;
end;

procedure inCash(bt: byte);		-- put byte in bf
begin
cash[wr]:=bt; inc(wr);
if wr=1000 then SaveCash;
end;

begin
wr:=0;

for y:=yw0 to yw1 do			-- compression is here (!)
 begin
 x:=xw0;
 repeat
 xGo:=x;
 i1:=getpixel(xGo,y);
  repeat
  inc(xGo);
  if xGo<xw1 then
   i2:=getpixel(xGo,y) else
   i2:=i1 xor 1;
  until (xGo-x>=j2) or (xGo>=xw1) or (i1<>i2);
 if (xGo-x>1) or (i1>=j1) then
  begin
  inCash(j1+xGo-x);
  inCash(i1); 
  end else 
  inCash(i1);
 x:=xGo;
 until x>=xw1;
 end;
if wr>0 then SaveCash;
end;

procedure _enc65536;			-- pcx 65536 encoder
var
cash		: array[0..999] of byte;
wr		: longint;
x,xGo,y		: word;
i1,i2		: byte;
pColor		: array[0..4095] of oneRGB;
plane		: word;
rt,gt,bt	: word;

procedure SaveCash;			-- save bf to disk
begin
blockwrite(f,cash,wr); wr:=0;
end;

procedure inCash(bt: byte);		-- put byte in bf
begin
cash[wr]:=bt; inc(wr);
if wr=1000 then SaveCash;
end;

begin
wr:=0;

for y:=yw0 to yw1 do			-- compression is here (!)
 begin
 for x:=xw0 to xw1 do			-- take all pixels of line
  with pColor[x] do			-- and remake it to RGB
   begin
   getRGBincs(getpixel(x,y),rt,gt,bt);
   r:=rt shl 2;				-- HLS stuff we need ...
   g:=gt shl 2;
   b:=bt shl 2;
   end;
 for plane:=0 to 2 do			-- now 3 planes present ...
 begin
 x:=xw0;
 repeat
 xGo:=x;
 case plane of
  0: i1:=pColor[xGo].r;
  1: i1:=pColor[xGo].g;
  2: i1:=pColor[xGo].b;
 end;
  repeat
  inc(xGo);
  if xGo<xw1 then
   begin
   case plane of
   0: i2:=pColor[xGo].r;
   1: i2:=pColor[xGo].g;
   2: i2:=pColor[xGo].b;
   end;
   end else
   i2:=i1 xor 1;
  until (xGo-x>=j2) or (xGo>=xw1) or (i1<>i2);
 if (xGo-x>1) or (i1>=j1) then
  begin
  inCash(j1+xGo-x);
  inCash(i1); 
  end else 
  inCash(i1);
 x:=xGo;
 until x>=xw1;
 end;
 end;
if wr>0 then SaveCash;
end;

begin

	maxColor:=getMaxColor;		-- take a max color

	if (xw0>xw1) or			-- check for a fool
	   (yw0>yw1) then
	begin
	PCXError:=notsupport;
	exit;
	end;

	if (maxColor<>255) and		-- check support
	   (maxColor<>65535) then
	begin
	PCXError:=notsupport;
	exit;
	end;

	with thisHead do		-- fill header
	begin
	manufact	:= 10;		-- zSoft
	version		:= 5;		-- 3.0
	encod		:= 1;		-- series
	bitpix		:= 8;
	x0		:= xw0;
	y0		:= yw0;
	x1		:= xw1;
	y1		:= yw1;
	hRes		:= xw1-xw0+1;
	vRes		:= yw1-yw0+1;
	reserved	:= 0;
	case MaxColor of		-- case for n-planes
 	  255 : NPlanes:=1;
	65535 : NPlanes:=3;
	end;
	bytesperline	:= xw1-xw0+1;
	palInfo		:= 0;
	fillchar(ColorMap,48,0);
	fillchar(Filler,58,0);
	end;

	assign(f,filename);		-- open future-PCX file
	rewrite(f,1);
	
	blockwrite(f,thisHead,128);	-- write Header

	inc(xw1);			-- special fun (!)

	case maxColor of		-- main encode
	  255 : _enc256;
	65535 : _enc65536;
	end;

	if maxColor=255 then		-- save a 256 hls pal
	 begin
	 for i:=0 to 767 do
	  tmppal[i]:=rgbpal[i] shl 2;
	 i:=pcxmagicbyte;
	 blockwrite(f,i,1);
	 blockwrite(f,tmppal,768);
	 end;

	close(f);			-- close with ok
	PCXError:=AllOk;

end;


begin
	fillChar(rgbpal,sizeof(rgbpal),0);
end.

